" Description: Omnicompletion hooks.
"              Also incorporates a small plugin that shows prototype
"              of a function in the message line.
" Maintainer:  Milan Straka
" Last Change: 03 Nov 2014

" Vim version 7.x is needed.
if v:version < 700
  echohl ErrorMsg | echomsg "completion_hooks.vim needs vim version >= 7.0!" | echohl None
  finish
endif

""""""""" Completion hooks

" Per-buffer, whether the implicit completion is allowed.
let b:ComplHookAllowImplicit = 1

" Per-buffer, whether the completion contains arguments.
let b:ComplHookWithArgs = 0

" Add reasonable mappings.
function ComplHookInit(initializers, quietAccepters, loudAccepters)
  let b:ComplHookAllowImplicit = 1
  let b:ComplHookWithArgs = 0

  for s:initializer in a:initializers
    if s:initializer == "\<C-X>\<C-O>"
      inoremap <expr> <buffer> <C-X><C-O> ComplHookStart(0, 0)
    elseif s:initializer == "\<C-X>\<C-A>"
      inoremap <expr> <buffer> <C-X><C-A> ComplHookStart(0, 1)
    else
      execute "inoremap <expr> <buffer>" s:initializer "ComplHookAccept('".s:initializer."').ComplHookStart(1)"
    endif
  endfor
  for s:loudAccepter in a:loudAccepters
    execute "inoremap <expr> <buffer>" s:loudAccepter "ComplHookAccept('".s:loudAccepter."')"
  endfor
  for s:quietAccepter in a:quietAccepters
    execute "inoremap <expr> <buffer>" s:quietAccepter "ComplHookAccept('".s:quietAccepter."', 1)"
  endfor

  setl completeopt-=menu
  setl completeopt+=menuone
  map <silent> <buffer> <C-X><C-B> :call ComplHookAllowImplicitToggle()<Cr>|map! <buffer> <C-X><C-B> <C-o><C-X><C-B>
  inoremap <expr> <buffer> <Esc><Space> ComplHookStart()
  inoremap <expr> <buffer> <Esc><Space> ComplHookStart()
  inoremap <expr> <silent> <buffer> <C-j> pumvisible() ? '<Down>' : '<C-j>'
  inoremap <expr> <silent> <buffer> <C-k> pumvisible() ? '<Up>' : '<C-k>'
endfun

" Init the completion in a reasonable way.
function ComplHookStart(...)
  let s:ComplHookImplicit = a:0 ? a:1 : 0
  if s:ComplHookImplicit && !b:ComplHookAllowImplicit
    return ""
  else
    let b:ComplHookWithArgs = a:0 >= 2 ? a:2 : 0
    call s:ComplHookOmnifuncWrap()
    return "\<C-X>\<C-O>\<C-P>\<Down>\<Down>"
  endif
endfun

function ComplHookAllowImplicitToggle()
  let b:ComplHookAllowImplicit = !b:ComplHookAllowImplicit
endfun

function s:ComplHookOmnifuncWrap()
  if &omnifunc != "ComplHookOmnifunc"
    let b:ComplHookOmnifuncOri = function(&omnifunc)
    setl omnifunc=ComplHookOmnifunc
  endif
endfun

function ComplHookOmnifunc(findstart, base)
  if a:findstart
    return b:ComplHookOmnifuncOri(a:findstart, a:base)
  else
    let s:completions = b:ComplHookOmnifuncOri(a:findstart, a:base)
    if b:ComplHookWithArgs
      for s:completion in s:completions
        let s:withArgs = has_key(s:completion, 'menu') ? s:completion.menu : has_key(s:completion, 'abbr') ? s:completion.abbr : ''
        if !empty(s:withArgs)
          let s:position = stridx(s:withArgs, s:completion.word)
          let s:completion.word = s:position > 0 ? strpart(s:withArgs, s:position) : s:withArgs
        endif
      endfor
    endif
    call complete_add({'word': a:base, 'empty': 1})
    return s:completions
  endif
endfun

" Accept selected completion. If it contains args and user pressed (, jump to
" the beginning of the arguments. If it contains args and user pressed ), do
" not repeat it. In any case, suppress <Cr> (we do not want to go to another line).
function ComplHookAccept(acceptedBy, ...)
  if !pumvisible()
    return a:acceptedBy
  else
    let s:ignore = a:0 ? a:1 : 0

    if b:ComplHookWithArgs && a:acceptedBy == '('
      return "\<C-y>\<C-o>T("
    elseif s:ignore || (b:ComplHookWithArgs && a:acceptedBy == ')')
      return "\<C-y>"
    else
      return "\<C-y>" . a:acceptedBy
    endif
  endif
endfun

""""""""" Show proto

" Enable show proto
function EnableShowProto()
  autocmd InsertLeave <buffer> call s:PrintNone()
  inoremap <silent> <buffer> <Esc>= <C-r>=ShowProto(1)<Cr>
  inoremap <silent> <buffer> <Esc>- <C-r>=ShowProto(-1)<Cr>
endfunction

fun ShowProto(delta)
  " initialize buffer variables if needed
  if !exists("b:showProtoPrototypes")|let b:showProtoPrototypes = []|endif
  if !exists("b:showProtoPosition")|let b:showProtoPosition = []|endif
  if !exists("b:showProtoCount")|let b:showProtoCount = 0|endif

  " save position of the cursor and move to end of desired function
  let s:cursor = getpos('.')
  let s:funFound = s:MoveToEndOfDesiredFunction()

  " find the prototypes and fill global structures if needed
  if s:funFound && getpos('.') != b:showProtoPosition
    call s:GetPrototypes()
    let b:showProtoPosition = getpos('.')
  elseif !s:funFound
    let b:showProtoPrototypes = []
    let b:showProtoPosition = []
  endif

  " advance and show the prototype if any
  if len(b:showProtoPrototypes)
    let b:showProtoCount += a:delta
    if b:showProtoCount < 0 | let b:showProtoCount=len(b:showProtoPrototypes)-1 | endif
    if b:showProtoCount >= len(b:showProtoPrototypes) | let b:showProtoCount=0 | endif
    call s:PrintProto()
  else
    call s:PrintNone()
  endif

  " move cursor back and return
  call cursor(s:cursor[1:])
  return ''
endfun

fun s:MoveToEndOfDesiredFunction()
  " we don't want to match ^\s*$ on the start line, adding \%<line(".")l to avoid
  let s:startLine = line(".")
  let s:brackets = 0
  while 1
    let s:what = search('\(^\%<'.s:startLine.'l\s*$\)\|\(;\)\|\((\)\|\()\)', 'bpW')
    "s:what is 0 if no match, 2 if blank line, 3 if ;, 4 if ( and 5 if )
    if s:what <= 3|return 0|endif
    if s:what==4 && s:brackets==0|break|endif
    let s:brackets += 2*s:what-9
  endwhile
  " found, move just one position left
  call cursor(line('.'), col('.')-1)
  return 1
endfun

fun s:GetPrototypes()
  let b:showProtoPrototypes = []
  let b:showProtoCount = -1
  let s:omnifunc = &omnifunc == "ComplHookOmnifunc" ? b:ComplHookOmnifuncOri : function(&omnifunc)
  let s:pos = getpos('.')
  let s:start = s:omnifunc(1, '')
  let s:name = getline('.')[(s:start):(s:pos[2]-1)]
  if &ft == 'python'|call setpos('.', [s:pos[0], s:pos[1], s:start+1, s:pos[3]])|endif
  for s:tagInfo in s:omnifunc(0, s:name)
    let s:prototype = has_key(s:tagInfo, 'info') ? s:tagInfo.info : ""
    let s:prototype = has_key(s:tagInfo, 'menu') && strlen(s:tagInfo.menu) > strlen(s:prototype) ? s:tagInfo.menu : s:prototype
    let s:prototype = has_key(s:tagInfo, 'abbr') && strlen(s:tagInfo.abbr) > strlen(s:prototype) ? s:tagInfo.abbr : s:prototype
    if empty(s:prototype) || (s:tagInfo.word != s:name && s:tagInfo.word != s:name."(")
      continue
    endif

    let s:piece = ''
    let s:piece_height = 0
    let s:piece_split = 0
    for line in split(s:prototype, '\n')
      if s:piece_height >= &lines / 2
        call add(b:showProtoPrototypes, s:piece)
        let s:piece = ''
        let s:piece_height = 0
        let s:piece_split = 1
      endif
      let s:piece_height += 1 + len(line) / &columns
      let s:piece .= line . "\n"
    endfor
    let s:piece = s:piece_split==1 ? s:piece . repeat("\n", &lines/2 - s:piece_height) : substitute(s:piece, '\n*$', '', '')
    call add(b:showProtoPrototypes, s:piece)

  endfor
  if &ft == 'python'|call setpos('.', s:pos)|endif
  unlet s:omnifunc

  " Number them if there are more
  if len(b:showProtoPrototypes)>1
    let s:i=0 | while s:i < len(b:showProtoPrototypes)
      let b:showProtoPrototypes[s:i] .= ' ('.(s:i+1).'/'.len(b:showProtoPrototypes).')'
      let s:i += 1
    endwhile
  endif
endfun

fun s:PrintProto()
  set noshowmode
  let s:msg = b:showProtoPrototypes[b:showProtoCount]
  let s:height = 0
  for line in split(s:msg, '\n', 1)
    let s:height += 1 + len(line) / &columns
  endfor
  let &cmdheight = s:height
  echo s:msg
endfun

fun s:PrintNone()
  set showmode cmdheight=1
  echo ''
endfun
